
def factory( name='collimator_radial',w1=0.015,h1=0.3,w2=0.015,h2=0.3,len=0.35,divergence=40.0,transmission=1.0,theta_min=5.0,theta_max=165.0,nchan=128.0,radius=0.0,nblades=0.0,roc=0.0,verbose=0.0 ):
    from mcstas2.bindings import boostpython
    from mcstas2.mcstas2bp import McStasComponentAsMcniComponent as component
    from Collimator_radialboostpython import Collimator_radial as _factory_mcvine
    return component( _factory_mcvine( name,w1,h1,w2,h2,len,divergence,transmission,theta_min,theta_max,nchan,radius,nblades,roc,verbose ) )

from mcstas2.utils.parsers.ComponentInfo import ComponentInfo, Parameter
info = ComponentInfo()
info_copyright = """Emmanuel Farhi <farhi@ill.fr>"""
info.copyright = info_copyright
info_declare = """
  double soller_theta1, soller_theta2, window_theta;
  char   use_approx;
  char   ideal     ;
"""
info.declare = info_declare
info_finalize = """"""
info.finalize = info_finalize
info_full_description = """Radial Soller collimator with rectangular opening and specified length.
The collimator is made of many rectangular channels stacked radially.
Each channel is a set of transmitting layers, separated by an absorbing
material, the whole stuff is inside an absorbing housing.
The component should be positioned at the radius center.
When specifying the divergence parameter, the transmission function is an
average (triangular approximation) and does not utilize knowledge of
the actual neutron trajectory.
On the other hand, when specifying the number of blades, model is exact,
but statistics is lowered by a factor 2.
A zero divergence disables collimation (then the component works as a double
slit). If the input/output (w1, w2) width are 0, then a perfect collimator
is assumed (no slit check except for the total angular range and height).
The component can be made oscillating (usual on diffractometers and TOF
machines) with the 'roc' parameter.
The neutron beam outside the collimator area is transmitted unaffected.

Ideal: Collimator_radial(name, w1=0, h1=.3, w2=0, h2=.3, len=.35,
divergence=40,transmission=1,
theta_min=5, theta_max=165, radius=0.9)"""
info.full_description = info_full_description
info_initialize = """
  use_approx=1;
  ideal     =0;
  double blades_thick;
  
  if (radius <= 0) exit(printf("Collimator_radial: %s: incorrect radius\n", NAME_CURRENT_COMP));
  if (nchan <= 0)  exit(printf("Collimator_radial: %s: incorrect number of channels\n", NAME_CURRENT_COMP));
  nchan = ceil(nchan);
  if (len <= 0)    exit(printf("Collimator_radial: %s: invalid collimator length\n", NAME_CURRENT_COMP));

  theta_max *= DEG2RAD;
  theta_min *= DEG2RAD;

  if (!w1 && !w2) { ideal=1; nblades=0; }
  else {
    soller_theta1  = atan2(fabs(w1), radius    );
    soller_theta2  = atan2(fabs(w2), radius+len);
    window_theta   = fabs(theta_max-theta_min)/nchan;

    if (soller_theta1 > window_theta || soller_theta2 > window_theta)
      exit(printf("Collimator_radial: %s: the %f channels of size %f [deg]\n"
                "                   do not fit within the angular range %g [deg]\n",
      NAME_CURRENT_COMP, nchan, soller_theta1*RAD2DEG, window_theta*RAD2DEG));
  }

  if (divergence && !nblades) {
    divergence *= MIN2RAD;
    blades_thick= len*tan(divergence);
    nblades     = fabs(floor(w1/blades_thick));
    nchan       =ceil(fabs(theta_max-theta_min)/divergence);
    window_theta=divergence;
    use_approx  = 1;
  } else if (nblades) {
    blades_thick = w1/nblades;
    divergence   = fabs(atan2(blades_thick,len));
    use_approx   = 0;
  } else { divergence = 0; nblades = 0; }
  if (verbose) {
    printf("Collimator_radial: %s: divergence %g [min] %s\n"
             "    Total opening [%.2g:%.2g] [deg] in %g channels",
             NAME_CURRENT_COMP, divergence*RAD2MIN,
             (roc ? "oscillating" : ""),
              theta_min*RAD2DEG, theta_max*RAD2DEG, nchan);
    if (!ideal)
      printf(" as %g channels of %.2g layers %s\n"
           "    Channel Window=%g [deg] Soller=%g [deg]\n",
           nchan, nblades, (use_approx ? "(approx)" : ""),
           window_theta*RAD2DEG, soller_theta1*RAD2DEG);
    else printf(" (ideal)\n");
  }

"""
info.initialize = info_initialize
item0 = Parameter()
item0_default = """collimator_radial"""
item0.default = item0_default
item0_description = """component name"""
item0.description = item0_description
item0_name = """name"""
item0.name = item0_name
item0_type = """char *"""
item0.type = item0_type
item1 = Parameter()
item1_default = 0.015
item1.default = item1_default
item1_description = """(m)          Input  window width"""
item1.description = item1_description
item1_name = """w1"""
item1.name = item1_name
item1_type = """double"""
item1.type = item1_type
item2 = Parameter()
item2_default = 0.3
item2.default = item2_default
item2_description = """(m)          Input  window height"""
item2.description = item2_description
item2_name = """h1"""
item2.name = item2_name
item2_type = """double"""
item2.type = item2_type
item3 = Parameter()
item3_default = 0.015
item3.default = item3_default
item3_description = """(m)          Output window width"""
item3.description = item3_description
item3_name = """w2"""
item3.name = item3_name
item3_type = """double"""
item3.type = item3_type
item4 = Parameter()
item4_default = 0.3
item4.default = item4_default
item4_description = """(m)          Output window height"""
item4.description = item4_description
item4_name = """h2"""
item4.name = item4_name
item4_type = """double"""
item4.type = item4_type
item5 = Parameter()
item5_default = 0.35
item5.default = item5_default
item5_description = """(m)          Length/Distance between slits"""
item5.description = item5_description
item5_name = """len"""
item5.name = item5_name
item5_type = """double"""
item5.type = item5_type
item6 = Parameter()
item6_default = 40.0
item6.default = item6_default
item6_description = """(min of arc) Divergence angle. May also be specified with the"""
item6.description = item6_description
item6_name = """divergence"""
item6.name = item6_name
item6_type = """double"""
item6.type = item6_type
item7 = Parameter()
item7_default = 1.0
item7.default = item7_default
item7_description = """(1)         Transmission of Soller (0<=t<=1)"""
item7.description = item7_description
item7_name = """transmission"""
item7.name = item7_name
item7_type = """double"""
item7.type = item7_type
item8 = Parameter()
item8_default = 5.0
item8.default = item8_default
item8_description = """(deg)        Minimum Theta angle for the radial setting"""
item8.description = item8_description
item8_name = """theta_min"""
item8.name = item8_name
item8_type = """double"""
item8.type = item8_type
item9 = Parameter()
item9_default = 165.0
item9.default = item9_default
item9_description = """(deg)        Maximum Theta angle for the radial setting"""
item9.description = item9_description
item9_name = """theta_max"""
item9.name = item9_name
item9_type = """double"""
item9.type = item9_type
item10 = Parameter()
item10_default = 128.0
item10.default = item10_default
item10_description = """(1)          Number of channels in the theta range"""
item10.description = item10_description
item10_name = """nchan"""
item10.name = item10_name
item10_type = """double"""
item10.type = item10_type
item11 = Parameter()
item11_default = 0.0
item11.default = item11_default
item11_description = """(m)          Radius of the collimator (to entry window)"""
item11.description = item11_description
item11_name = """radius"""
item11.name = item11_name
item11_type = """double"""
item11.type = item11_type
item12 = Parameter()
item12_default = 0.0
item12.default = item12_default
item12_description = """(1)         Number of blades composing each Soller. Overrides"""
item12.description = item12_description
item12_name = """nblades"""
item12.name = item12_name
item12_type = """double"""
item12.type = item12_type
item13 = Parameter()
item13_default = 0.0
item13.default = item13_default
item13_description = """(1)         Enable oscillation of collimator with an amplitude"""
item13.description = item13_description
item13_name = """roc"""
item13.name = item13_name
item13_type = """double"""
item13.type = item13_type
item14 = Parameter()
item14_default = 0.0
item14.default = item14_default
item14_description = """(0/1)       Gives additional information"""
item14.description = item14_description
item14_name = """verbose"""
item14.name = item14_name
item14_type = """double"""
item14.type = item14_type
info_input_parameters = [ item0, item1, item2, item3, item4, item5, item6, item7, item8, item9, item10, item11, item12, item13, item14 ]
info.input_parameters = info_input_parameters
info_name = """Collimator_radial"""
info.name = info_name
item0 = Parameter()
item0_default = 0
item0.default = item0_default
item0_description = """"""
item0.description = item0_description
item0_name = """soller_theta1"""
item0.name = item0_name
item0_type = """double"""
item0.type = item0_type
item1 = Parameter()
item1_default = 0
item1.default = item1_default
item1_description = """"""
item1.description = item1_description
item1_name = """soller_theta2"""
item1.name = item1_name
item1_type = """double"""
item1.type = item1_type
item2 = Parameter()
item2_default = 0
item2.default = item2_default
item2_description = """"""
item2.description = item2_description
item2_name = """window_theta"""
item2.name = item2_name
item2_type = """double"""
item2.type = item2_type
item3 = Parameter()
item3_default = 0
item3.default = item3_default
item3_description = """"""
item3.description = item3_description
item3_name = """use_approx"""
item3.name = item3_name
item3_type = """double"""
item3.type = item3_type
item4 = Parameter()
item4_default = 0
item4.default = item4_default
item4_description = """"""
item4.description = item4_description
item4_name = """ideal"""
item4.name = item4_name
item4_type = """double"""
item4.type = item4_type
info_output_parameters = [ item0, item1, item2, item3, item4 ]
info.output_parameters = info_output_parameters
info_save = """"""
info.save = info_save
item0 = """"""
item1 = """"""
info_share = [ item0, item1 ]
info.share = info_share
info_simple_description = """A radial Soller collimator."""
info.simple_description = info_simple_description
item0 = """x"""
item1 = """y"""
item2 = """z"""
item3 = """vx"""
item4 = """vy"""
item5 = """vz"""
item6 = """t"""
item7 = """s1"""
item8 = """s2"""
item9 = """p"""
info_state_parameters = [ item0, item1, item2, item3, item4, item5, item6, item7, item8, item9 ]
info.state_parameters = info_state_parameters
info_trace = """
  double phi, t0, t1, t2, t3;
  int    intersect;
  long   input_chan,  output_chan;
  long   input_blade, output_blade;
  double input_theta, output_theta;
  double input_center,output_center;
  char   ok=0;
  double roc_theta=0;

  
  intersect=cylinder_intersect(&t0,&t3,x,y,z,vx,vy,vz,radius,h1);
  if (!intersect) ABSORB;
  else if (t3 > t0) t0 = t3;

  intersect=cylinder_intersect(&t1,&t2,x,y,z,vx,vy,vz,radius+len,h2);
  if (!intersect) ABSORB;
  else if (t2 > t1) t1 = t2;

  
  if (t0 > 0 && t1 > t0) {
    PROP_DT(t0);
    if (!ideal) {
      if (roc) roc_theta = roc*randpm1()*soller_theta1/2;

      input_theta = atan2(x, z) + roc_theta;
    
      input_chan  = floor((input_theta - theta_min)/window_theta);
      if (input_chan >= 0 && input_chan < nchan && fabs(y) <= h1/2) ok=1;
    } else if (fabs(y) < h1/2) ok = 1;

    if (ok) {
      if (ideal)
        input_center=atan2(x,z);
      else {
        input_center= theta_min + input_chan*window_theta
                  + (window_theta)/2;
        
        phi = input_theta - input_center;

        if (fabs(phi) > soller_theta1/2) ABSORB; 
      
        input_blade = floor(phi/soller_theta1*nblades);
      }
      SCATTER;

      
      PROP_DT(t1-t0);
      SCATTER;
      if (!ideal) {
        output_theta = atan2(x, z) + roc_theta;
        
        output_chan  = floor((output_theta - theta_min)/window_theta);
        
        if (output_chan != input_chan) ABSORB; 
        output_center= theta_min + output_chan*window_theta
                    + (window_theta)/2;
        
        phi = output_theta -output_center;
        if (fabs(phi) > soller_theta2/2 || fabs(y) > h2/2) ABSORB; 
        
        output_blade = floor(phi/soller_theta2*nblades);
      } else if (fabs(y) > h2/2) ABSORB;

      
      if (divergence) {
        if (use_approx)
        {
          phi = fabs(atan2(vx,vz) - input_center); 
          if (phi > divergence)
            ABSORB; 
          else
            p *= transmission*(1.0 - phi/divergence);
          SCATTER;
        } else {
          if (input_blade != output_blade) ABSORB; 
          p *= transmission;
          SCATTER;
        }
      }
    } 
  } 

"""
info.trace = info_trace

factory.info = info
factory.__doc__ = str(info)
